/*
 * serial_qt.h - A driver for Quatech's asychronous PCMCIA serial devices
 * Copyright (C) 2001 Tim Gobeli - Quatech Inc.

 * this program is free software; you cand redistribute it and/or modify 
 * it under the terms of the GNU General Public License as  published by 
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.

 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Publice License for more details.

 *  serial_qt.h 1.00 2001/10/29 9:40:30
 *
 * The initial developer of the original code is David A. Hinds
 * <dahinds@users.sourceforge.net>.  Portions created by David A. Hinds
 * are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.
 *

 */
 
//Uncomment for debugging
//#define DEBUG_ON

#ifdef DEBUG_ON
    #define  dbgprint(const...)    printk(const)

#else
    #define  dbgprint(const...)
#endif
 

#ifndef _LINUX_CISCODE_H
#define _LINUX_CISCODE_H

/* Manufacturer and Product ID codes */


#define MANFID_QUATECH			0x0137
#define PRODID_QUATECH_SINGLE_RS232	0x0001
#define PRODID_QUATECH_SINGLE_RS232_G	0x004b
#define PRODID_QUATECH_SINGLE_RS422	0x0002
#define PRODID_QUATECH_SINGLE_RS422_G	0x004c

#define PRODID_QUATECH_DUAL_RS232	0x0007
#define PRODID_QUATECH_DUAL_RS232_G	0x004d
#define PRODID_QUATECH_DUAL_RS422	0x000e
#define PRODID_QUATECH_DUAL_RS422_G     0x004e


#define PRODID_QUATECH_QUAD_RS232	0x001b
#define PRODID_QUATECH_QUAD_RS422	0x0045


#define SERIALQT_PCI_IOC_MAGIC 'k'
#define SERIALQT_WRITE_QOPR _IOW(SERIALQT_PCI_IOC_MAGIC, 0, int)
#define SERIALQT_WRITE_QMCR _IOW(SERIALQT_PCI_IOC_MAGIC, 1, int)
#define SERIALQT_GET_NUMOF_UNITS _IOR(SERIALQT_PCI_IOC_MAGIC, 2, void *)
#define SERIALQT_GET_THIS_UNIT _IOR(SERIALQT_PCI_IOC_MAGIC, 3, void *)
#define SERIALQT_READ_QOPR _IOR(SERIALQT_PCI_IOC_MAGIC, 4, int)
#define SERIALQT_READ_QMCR _IOR(SERIALQT_PCI_IOC_MAGIC, 5, int)
#define SERIALQT_IS422_EXTENDED _IOR(SERIALQT_PCI_IOC_MAGIC, 6, int)    //returns successful if 422 extended


#define SERIALQT_IOC_MAXNR 6

 // --------------------------------------------------------------------------------------------------------------------
// Constants to Check for presence fo QOR register
// --------------------------------------------------------------------------------------------------------------------


#define QPCR_TEST_FOR1    0x3f  // 
#define QPCR_TEST_GET1    0x00  // (0 0 x x x x x x) in gives (0 0 x x x x x x) out
#define QPCR_TEST_FOR2    0x40  // 
#define QPCR_TEST_GET2    0x040  // (0 1 x x x x x x) in gives (0 1 x x x x x x) out
#define QPCR_TEST_FOR3    0x80  // 
#define QPCR_TEST_GET3    0x40  // (1 0 x x x x x x) in gives (0 1 x x x x x x) out
#define QPCR_TEST_FOR4    0xc0  // 
#define QPCR_TEST_GET4    0x80  // (1 1 x x x x x x) in gives (1 0 x x x x x x) out

// --------------------------------------------------------------------------------------------------------------------
// Quatech Options Register DWORD "QOPR" (Hardware Register)
// --------------------------------------------------------------------------------------------------------------------
// Lower 8 bits defined by hardware registers, upper 8 bits custom flags
#define QOPR_CLOCK_AUTO    0x0100  // (bit 8) custom flag for "Auto" data rate multiplier mode
#define QOPR_CLOCK_X1      0x0000  // (x x x x x x 0 0) force X1 clock mode   
#define QOPR_CLOCK_X2      0x0001  // (x x x x x x 0 1) force X2 clock mode
#define QOPR_CLOCK_X4      0x0002  // (x x x x x x 1 0) force X4 clock mode
#define QOPR_CLOCK_X8      0x0003  // (x x x x x x 1 1) force X8 clock mode
#define QOPR_CLOCK_MASK    0x0103  // (1)(x x x x x x 1 1) used by software to mask off bits
#define QOPR_CLOCK_RATE_MASK    0x0003  // (1)(x x x x x x 1 1) used by software to mask off bits

// -----------------------------------------------------------------------------------------------
// low word, low byte in HWOR defines clock multiplier options available 
#define CLOCK_X1           115200  //  X1 clock RATE
#define CLOCK_X2           230400  // X2 clock RATE
#define CLOCK_X4           460800  // X4 clock RATE
#define CLOCK_X8           921600  // (X8 clock RATE

#define MULTI_COUNT (sizeof(multi_id)/sizeof(multi_id_t))



#define TRUE 1
#define FALSE 0
#define bool int

typedef struct
{
    int     index;
    int     n_identity;           
            
} Identity_struct;


/* Table of multi-port card ID's */

typedef struct
{
    u_short manfid;
    u_short prodid;
    int     multi;      /* 1 = multifunction, > 1 = # ports */
    int         extended;         /* 1 = extended feature, 1 = no extended feature */
} multi_id_t;



static multi_id_t multi_id[] = {
    { MANFID_QUATECH, PRODID_QUATECH_SINGLE_RS232, 1, 0},
    { MANFID_QUATECH, PRODID_QUATECH_SINGLE_RS232_G, 1, 1},
    { MANFID_QUATECH, PRODID_QUATECH_SINGLE_RS422, 1, 0},
    { MANFID_QUATECH, PRODID_QUATECH_SINGLE_RS422_G, 1, 1},
    { MANFID_QUATECH, PRODID_QUATECH_DUAL_RS232, 2, 0},
    { MANFID_QUATECH, PRODID_QUATECH_DUAL_RS232_G, 2, 1},
    { MANFID_QUATECH, PRODID_QUATECH_DUAL_RS422, 2, 0},
    { MANFID_QUATECH, PRODID_QUATECH_DUAL_RS422_G, 2, 1},
    { MANFID_QUATECH, PRODID_QUATECH_QUAD_RS232, 4, 0},
    { MANFID_QUATECH, PRODID_QUATECH_QUAD_RS422, 4, 0},
};


typedef struct serial_info_t
{
    dev_link_t  link;
    int     ndev;
    int     multi;
    int     slave;
    int     manfid;
    int     prodid;
    dev_node_t  node[4];
    int     line[4];
    int         extended;
    int     Adapter_Index;
    int base_major_number;

} serial_info_t;



/*****************functions ************************************/
static void serial_release(u_long arg);
static int serial_event(event_t event, int priority,
                        event_callback_args_t *args);

static dev_info_t dev_info = "serial_qt";

static dev_link_t *serial_attach(void);
static void serial_detach_one(u_long);

static void serial_detach(dev_link_t *link);


static void serial_config(dev_link_t *link);

static int ioctl_serial_pcmcia (struct inode *innod, struct file *filp, unsigned int cmd,
                             unsigned long arg);

static unsigned SerialReadQOPR_Reg(unsigned Base_Address);

static void SerialWriteQOPR_Reg(unsigned Base_Address, bool internal, unsigned Arg);

static void SerialWriteQMCR_Reg(unsigned Base_Address, unsigned Arg);

static unsigned SerialReadQMCR_Reg(unsigned Base_Address);

static int SerialDoes422_Exist(unsigned Base_Address);

static int SerialDoesQMCR_Exist(unsigned Base_Address);


#endif /* _LINUX_CISCODE_H */
